@@ -9,6 +9,7 @@ import time
 from contextlib import contextmanager
 from copy import deepcopy
 from datetime import datetime
+from enum import IntEnum
 from threading import current_thread, Lock
 from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple, TYPE_CHECKING, Union
 
@@ -17,7 +18,7 @@ from psutil import TimeoutExpired
 
 from .. import global_config, psycopg
 from ..async_executor import CriticalTask
-from ..collections import CaseInsensitiveDict, EMPTY_DICT
+from ..collections import CaseInsensitiveDict, CaseInsensitiveSet, EMPTY_DICT
 from ..dcs import Cluster, Leader, Member, slot_name_from_member_name
 from ..exceptions import PostgresConnectionException
 from ..tags import Tags
@@ -27,7 +28,7 @@ from .callback_executor import CallbackAction, CallbackExecutor
 from .cancellable import CancellableSubprocess
 from .config import ConfigHandler, mtime
 from .connection import ConnectionPool, get_connection_cursor
-from .misc import parse_history, parse_lsn, postgres_major_version_to_int
+from .misc import parse_history, parse_lsn, postgres_major_version_to_int, PostgresqlRole, PostgresqlState
 from .mpp import AbstractMPP
 from .postmaster import PostmasterProcess
 from .slots import SlotsHandler
@@ -39,14 +40,24 @@ if TYPE_CHECKING:  # pragma: no cover
 
 logger = logging.getLogger(__name__)
 
-STATE_RUNNING = 'running'
-STATE_REJECT = 'rejecting connections'
-STATE_NO_RESPONSE = 'not responding'
-STATE_UNKNOWN = 'unknown'
-
 STOP_POLLING_INTERVAL = 1
 
 
+class PgIsReadyStatus(IntEnum):
+    """Possible PostgreSQL connection status ``pg_isready`` utility can report.
+
+    :cvar RUNNING: return code 0. PostgreSQL is accepting connections normally.
+    :cvar REJECT: return code 1. PostgreSQL is rejecting connections.
+    :cvar NO_RESPONSE: return code 2. There was no response to the connection attempt.
+    :cvar UNKNOWN: Return code 3. No connection attempt was made, something went wrong.
+    """
+
+    RUNNING = 0
+    REJECT = 1
+    NO_RESPONSE = 2
+    UNKNOWN = 3
+
+
 @contextmanager
 def null_context():
     yield
@@ -76,13 +87,15 @@ class Postgresql(object):
         self._major_version = self.get_major_version()
 
         self._state_lock = Lock()
-        self.set_state('stopped')
+        self.set_state(PostgresqlState.STOPPED)
 
         self._pending_restart_reason = CaseInsensitiveDict()
         self.connection_pool = ConnectionPool()
         self._connection = self.connection_pool.get('heartbeat')
         self.mpp_handler = mpp.get_handler_impl(self)
         self._bin_dir = config.get('bin_dir') or ''
+        self._role_lock = Lock()
+        self.set_role(PostgresqlRole.UNINITIALIZED)
         self.config = ConfigHandler(self, config)
         self.config.check_directories()
 
@@ -107,7 +120,6 @@ class Postgresql(object):
         self._is_leader_retry = Retry(max_tries=1, deadline=config['retry_timeout'] / 2.0, max_delay=1,
                                       retry_exceptions=PostgresConnectionException)
 
-        self._role_lock = Lock()
         self.set_role(self.get_postgres_role_from_data_directory())
         self._state_entry_timestamp = 0
 
@@ -119,29 +131,31 @@ class Postgresql(object):
         # Last known running process
         self._postmaster_proc = None
 
+        self._available_gucs = None
+
         if self.is_running():
             # If we found postmaster process we need to figure out whether postgres is accepting connections
-            self.set_state('starting')
+            self.set_state(PostgresqlState.STARTING)
             self.check_startup_state_changed()
 
-        if self.state == 'running':  # we are "joining" already running postgres
+        if self.state == PostgresqlState.RUNNING:  # we are "joining" already running postgres
             # we know that PostgreSQL is accepting connections and can read some GUC's from pg_settings
             self.config.load_current_server_parameters()
 
-            self.set_role('primary' if self.is_primary() else 'replica')
+            self.set_role(PostgresqlRole.PRIMARY if self.is_primary() else PostgresqlRole.REPLICA)
 
             hba_saved = self.config.replace_pg_hba()
             ident_saved = self.config.replace_pg_ident()
 
-            if self.major_version < 120000 or self.role == 'primary':
+            if self.major_version < 120000 or self.role == PostgresqlRole.PRIMARY:
                 # If PostgreSQL is running as a primary or we run PostgreSQL that is older than 12 we can
                 # call reload_config() once again (the first call happened in the ConfigHandler constructor),
                 # so that it can figure out if config files should be updated and pg_ctl reload executed.
                 self.config.reload_config(config, sighup=bool(hba_saved or ident_saved))
             elif hba_saved or ident_saved:
                 self.reload()
-        elif not self.is_running() and self.role == 'primary':
-            self.set_role('demoted')
+        elif not self.is_running() and self.role == PostgresqlRole.PRIMARY:
+            self.set_role(PostgresqlRole.DEMOTED)
 
     @property
     def create_replica_methods(self) -> List[str]:
@@ -223,26 +237,37 @@ class Postgresql(object):
                          " pg_catalog.pg_stat_get_activity(w.pid)"
                          " WHERE w.state = 'streaming') r)").format(self.wal_name, self.lsn_name)
                         if global_config.is_synchronous_mode
-                        and self.role in ('primary', 'promoted') else "'on', '', NULL")
+                        and self.role in (PostgresqlRole.PRIMARY, PostgresqlRole.PROMOTED) else "'on', '', NULL")
 
         if self._major_version >= 90600:
+            filter_failover = ' WHERE NOT failover' if self._major_version >= 170000 else ''
             extra = ("pg_catalog.current_setting('restore_command')" if self._major_version >= 120000 else "NULL") +\
                 ", " + ("(SELECT pg_catalog.json_agg(s.*) FROM (SELECT slot_name, slot_type as type, datoid::bigint, "
                         "plugin, catalog_xmin, pg_catalog.pg_wal_lsn_diff(confirmed_flush_lsn, '0/0')::bigint"
                         " AS confirmed_flush_lsn, pg_catalog.pg_wal_lsn_diff(restart_lsn, '0/0')::bigint"
-                        " AS restart_lsn FROM pg_catalog.pg_get_replication_slots()) AS s)"
+                        f" AS restart_lsn, xmin FROM pg_catalog.pg_get_replication_slots(){filter_failover}) AS s)"
                         if self._should_query_slots and self.can_advance_slots else "NULL") + extra
-            extra = (", CASE WHEN latest_end_lsn IS NULL THEN NULL ELSE received_tli END,"
-                     " slot_name, conninfo, status, {0} FROM pg_catalog.pg_stat_get_wal_receiver()").format(extra)
-            if self.role == 'standby_leader':
+
+            written_lsn = ("pg_catalog.pg_wal_lsn_diff(written_lsn, '0/0')::bigint"
+                           if self._major_version >= 130000 else "NULL")
+            extra = (", CASE WHEN latest_end_lsn IS NULL THEN NULL ELSE received_tli END, {0}, slot_name, "
+                     "conninfo, status, {1} FROM pg_catalog.pg_stat_get_wal_receiver()").format(written_lsn, extra)
+            if self.role == PostgresqlRole.STANDBY_LEADER:
                 extra = "timeline_id" + extra + ", pg_catalog.pg_control_checkpoint()"
             else:
                 extra = "0" + extra
         else:
-            extra = "0, NULL, NULL, NULL, NULL, NULL, NULL" + extra
+            extra = "0, NULL, NULL, NULL, NULL, NULL, NULL, NULL" + extra
 
         return ("SELECT " + self.TL_LSN + ", {3}").format(self.wal_name, self.lsn_name, self.wal_flush, extra)
 
+    @property
+    def available_gucs(self) -> CaseInsensitiveSet:
+        """GUCs available in this Postgres server."""
+        if not self._available_gucs:
+            self._available_gucs = self._get_gucs()
+        return self._available_gucs
+
     def _version_file_exists(self) -> bool:
         return not self.data_directory_empty() and os.path.isfile(self._version_file)
 
@@ -290,10 +315,10 @@ class Postgresql(object):
         initdb = [self.pgcommand('initdb')] + list(args) + [self.data_dir]
         return subprocess.call(initdb, **kwargs) == 0
 
-    def pg_isready(self) -> str:
+    def pg_isready(self) -> PgIsReadyStatus:
         """Runs pg_isready to see if PostgreSQL is accepting connections.
 
-        :returns: 'ok' if PostgreSQL is up, 'reject' if starting up, 'no_resopnse' if not up."""
+        :returns: one of :class:`PgIsReadyStatus` values."""
 
         r = self.connection_pool.conn_kwargs
         cmd = [self.pgcommand('pg_isready'), '-p', r['port'], '-d', self._database]
@@ -307,11 +332,10 @@ class Postgresql(object):
             cmd.extend(['-U', r['user']])
 
         ret = subprocess.call(cmd)
-        return_codes = {0: STATE_RUNNING,
-                        1: STATE_REJECT,
-                        2: STATE_NO_RESPONSE,
-                        3: STATE_UNKNOWN}
-        return return_codes.get(ret, STATE_UNKNOWN)
+        try:
+            return PgIsReadyStatus(ret)
+        except ValueError:
+            return PgIsReadyStatus.UNKNOWN
 
     def reload_config(self, config: Dict[str, Any], sighup: bool = False) -> None:
         self.config.reload_config(config, sighup)
@@ -342,13 +366,13 @@ class Postgresql(object):
             self._sysid = data.get('Database system identifier', '')
         return self._sysid
 
-    def get_postgres_role_from_data_directory(self) -> str:
+    def get_postgres_role_from_data_directory(self) -> PostgresqlRole:
         if self.data_directory_empty() or not self.controldata():
-            return 'uninitialized'
+            return PostgresqlRole.UNINITIALIZED
         elif self.config.recovery_conf_exists():
-            return 'replica'
+            return PostgresqlRole.REPLICA
         else:
-            return 'primary'
+            return PostgresqlRole.PRIMARY
 
     @property
     def server_version(self) -> int:
@@ -375,7 +399,7 @@ class Postgresql(object):
         try:
             return self._connection.query(sql, *params)
         except PostgresConnectionException as exc:
-            if self.state == 'restarting':
+            if self.state == PostgresqlState.RESTARTING:
                 raise RetryFailedError('cluster is being restarted') from exc
             raise
 
@@ -465,10 +489,10 @@ class Postgresql(object):
         if not self._cluster_info_state:
             try:
                 result = self._is_leader_retry(self._query, self.cluster_info_query)[0]
-                cluster_info_state = dict(zip(['timeline', 'wal_position', 'replayed_location',
-                                               'received_location', 'replay_paused', 'pg_control_timeline',
-                                               'received_tli', 'slot_name', 'conninfo', 'receiver_state',
-                                               'restore_command', 'slots', 'synchronous_commit',
+                cluster_info_state = dict(zip(['timeline', 'wal_position', 'replay_lsn',
+                                               'receive_lsn', 'replay_paused', 'pg_control_timeline',
+                                               'received_tli', 'write_location', 'slot_name', 'conninfo',
+                                               'receiver_state', 'restore_command', 'slots', 'synchronous_commit',
                                                'synchronous_standby_names', 'pg_stat_replication'], result))
                 if self._should_query_slots and self.can_advance_slots:
                     cluster_info_state['slots'] =\
@@ -476,19 +500,21 @@ class Postgresql(object):
                 self._cluster_info_state = cluster_info_state
             except RetryFailedError as e:  # SELECT failed two times
                 self._cluster_info_state = {'error': str(e)}
-                if not self.is_starting() and self.pg_isready() == STATE_REJECT:
-                    self.set_state('starting')
+                if not self.is_starting() and self.pg_isready() == PgIsReadyStatus.REJECT:
+                    self.set_state(PostgresqlState.STARTING)
 
         if 'error' in self._cluster_info_state:
             raise PostgresConnectionException(self._cluster_info_state['error'])
 
         return self._cluster_info_state.get(name)
 
-    def replayed_location(self) -> Optional[int]:
-        return self._cluster_info_state_get('replayed_location')
+    def replay_lsn(self) -> Optional[int]:
+        return self._cluster_info_state_get('replay_lsn')
 
-    def received_location(self) -> Optional[int]:
-        return self._cluster_info_state_get('received_location')
+    def receive_lsn(self) -> Optional[int]:
+        write = self._cluster_info_state_get('write_location')
+        received = self._cluster_info_state_get('receive_lsn')
+        return max(received, write) if received and write else write or received
 
     def slots(self) -> Dict[str, int]:
         """Get replication slots state.
@@ -531,7 +557,7 @@ class Postgresql(object):
         """Figure out the replication state from input parameters.
 
         .. note::
-            This method could be only called when Postgres is up, running and queries are successfuly executed.
+            This method could be only called when Postgres is up, running and queries are successfully executed.
 
         :is_primary: `True` is postgres is not running in recovery
         :receiver_state: value from `pg_stat_get_wal_receiver.state` or None if Postgres is older than 9.6
@@ -566,7 +592,7 @@ class Postgresql(object):
             return bool(self._cluster_info_state_get('timeline'))
         except PostgresConnectionException:
             logger.warning('Failed to determine PostgreSQL state from the connection, falling back to cached role')
-            return bool(self.is_running() and self.role == 'primary')
+            return bool(self.is_running() and self.role == PostgresqlRole.PRIMARY)
 
     def replay_paused(self) -> bool:
         return self._cluster_info_state_get('replay_paused') or False
@@ -597,7 +623,7 @@ class Postgresql(object):
                 return match.group(1), match.group(2), match.group(3), match.group(4)
         return None, None, None, None
 
-    def _checkpoint_locations_from_controldata(self, data: Dict[str, str]) -> Optional[Tuple[int, int]]:
+    def latest_checkpoint_locations(self, data: Optional[Dict[str, str]] = None) -> Tuple[Optional[int], Optional[int]]:
         """Get shutdown checkpoint location.
 
         :param data: :class:`dict` object with values returned by `pg_controldata` tool.
@@ -605,6 +631,8 @@ class Postgresql(object):
         :returns: a tuple of checkpoint LSN for the cleanly shut down primary, and LSN of prev wal record (SWITCH)
                   if we know that the checkpoint was written to the new WAL file due to the archive_mode=on.
         """
+        if data is None:
+            data = self.controldata()
         timeline = data.get("Latest checkpoint's TimeLineID")
         lsn = checkpoint_lsn = data.get('Latest checkpoint location')
         prev_lsn = None
@@ -625,19 +653,7 @@ class Postgresql(object):
                 logger.error('Exception when parsing WAL pg_%sdump output: %r', self.wal_name, e)
             if isinstance(checkpoint_lsn, int):
                 return checkpoint_lsn, (prev_lsn or checkpoint_lsn)
-
-    def latest_checkpoint_location(self) -> Optional[int]:
-        """Get shutdown checkpoint location.
-
-        .. note::
-            In case if checkpoint was written to the new WAL file due to the archive_mode=on
-            we return LSN of the previous wal record (SWITCH).
-
-        :returns: checkpoint LSN for the cleanly shut down primary.
-        """
-        checkpoint_locations = self._checkpoint_locations_from_controldata(self.controldata())
-        if checkpoint_locations:
-            return checkpoint_locations[1]
+        return None, None
 
     def is_running(self) -> Optional[PostmasterProcess]:
         """Returns PostmasterProcess if one is running on the data directory or None. If most recently seen process
@@ -646,6 +662,7 @@ class Postgresql(object):
             if self._postmaster_proc.is_running():
                 return self._postmaster_proc
             self._postmaster_proc = None
+            self._available_gucs = None
 
         # we noticed that postgres was restarted, force syncing of replication slots and check of logical slots
         self.slots_handler.schedule()
@@ -667,7 +684,7 @@ class Postgresql(object):
 
         if self.callback and cb_type in self.callback:
             cmd = self.callback[cb_type]
-            role = 'primary' if self.role == 'promoted' else self.role
+            role = PostgresqlRole.PRIMARY if self.role == PostgresqlRole.PROMOTED else self.role
             try:
                 cmd = shlex.split(self.callback[cb_type]) + [cb_type, role, self.scope]
                 self._callback_executor.call(cmd)
@@ -675,20 +692,20 @@ class Postgresql(object):
                 logger.exception('callback %s %r %s %s failed', cmd, cb_type, role, self.scope)
 
     @property
-    def role(self) -> str:
+    def role(self) -> PostgresqlRole:
         with self._role_lock:
             return self._role
 
-    def set_role(self, value: str) -> None:
+    def set_role(self, value: PostgresqlRole) -> None:
         with self._role_lock:
             self._role = value
 
     @property
-    def state(self) -> str:
+    def state(self) -> PostgresqlState:
         with self._state_lock:
             return self._state
 
-    def set_state(self, value: str) -> None:
+    def set_state(self, value: PostgresqlState) -> None:
         with self._state_lock:
             self._state = value
             self._state_entry_timestamp = time.time()
@@ -697,7 +714,7 @@ class Postgresql(object):
         return time.time() - self._state_entry_timestamp
 
     def is_starting(self) -> bool:
-        return self.state == 'starting'
+        return self.state in (PostgresqlState.STARTING, PostgresqlState.BOOTSTRAP_STARTING)
 
     def wait_for_port_open(self, postmaster: PostmasterProcess, timeout: float) -> bool:
         """Waits until PostgreSQL opens ports."""
@@ -707,12 +724,12 @@ class Postgresql(object):
 
             if not postmaster.is_running():
                 logger.error('postmaster is not running')
-                self.set_state('start failed')
+                self.set_state(PostgresqlState.START_FAILED)
                 return False
 
             isready = self.pg_isready()
-            if isready != STATE_NO_RESPONSE:
-                if isready not in [STATE_REJECT, STATE_RUNNING]:
+            if isready != PgIsReadyStatus.NO_RESPONSE:
+                if isready not in [PgIsReadyStatus.REJECT, PgIsReadyStatus.RUNNING]:
                     logger.warning("Can't determine PostgreSQL startup status, assuming running")
                 return True
 
@@ -720,7 +737,7 @@ class Postgresql(object):
         return False
 
     def start(self, timeout: Optional[float] = None, task: Optional[CriticalTask] = None,
-              block_callbacks: bool = False, role: Optional[str] = None,
+              block_callbacks: bool = False, role: Optional[PostgresqlRole] = None,
               after_start: Optional[Callable[..., Any]] = None) -> Optional[bool]:
         """Start PostgreSQL
 
@@ -735,9 +752,11 @@ class Postgresql(object):
         # patroni.
         self.connection_pool.close()
 
+        state = (PostgresqlState.BOOTSTRAP_STARTING if self.bootstrap.running_custom_bootstrap
+                 else PostgresqlState.STARTING)
         if self.is_running():
             logger.error('Cannot start PostgreSQL because one is already running.')
-            self.set_state('starting')
+            self.set_state(state)
             return True
 
         if not block_callbacks:
@@ -745,7 +764,7 @@ class Postgresql(object):
 
         self.set_role(role or self.get_postgres_role_from_data_directory())
 
-        self.set_state('starting')
+        self.set_state(state)
         self.set_pending_restart_reason(CaseInsensitiveDict())
 
         try:
@@ -820,7 +839,7 @@ class Postgresql(object):
                 cur.execute('CHECKPOINT')
         except psycopg.Error:
             logger.exception('Exception during CHECKPOINT')
-            return 'not accessible or not healty'
+            return 'not accessible or not healthy'
 
     def stop(self, mode: str = 'fast', block_callbacks: bool = False, checkpoint: Optional[bool] = None,
              on_safepoint: Optional[Callable[..., Any]] = None, on_shutdown: Optional[Callable[[int, int], Any]] = None,
@@ -844,12 +863,12 @@ class Postgresql(object):
             # block_callbacks is used during restart to avoid
             # running start/stop callbacks in addition to restart ones
             if not block_callbacks:
-                self.set_state('stopped')
+                self.set_state(PostgresqlState.STOPPED)
                 if pg_signaled:
                     self.call_nowait(CallbackAction.ON_STOP)
         else:
             logger.warning('pg_ctl stop failed')
-            self.set_state('stop failed')
+            self.set_state(PostgresqlState.STOP_FAILED)
         return success
 
     def _do_stop(self, mode: str, block_callbacks: bool, checkpoint: bool,
@@ -865,7 +884,7 @@ class Postgresql(object):
             self.checkpoint(timeout=stop_timeout)
 
         if not block_callbacks:
-            self.set_state('stopping')
+            self.set_state(PostgresqlState.STOPPING)
 
         # invoke user-directed before stop script
         self._before_stop()
@@ -893,9 +912,9 @@ class Postgresql(object):
             while postmaster.is_running():
                 data = self.controldata()
                 if data.get('Database cluster state', '') == 'shut down':
-                    checkpoint_locations = self._checkpoint_locations_from_controldata(data)
-                    if checkpoint_locations:
-                        on_shutdown(*checkpoint_locations)
+                    checkpoint_lsn, prev_lsn = self.latest_checkpoint_locations(data)
+                    if checkpoint_lsn is not None and prev_lsn is not None:
+                        on_shutdown(checkpoint_lsn, prev_lsn)
                     break
                 elif data.get('Database cluster state', '').startswith('shut down'):  # shut down in recovery
                     break
@@ -955,28 +974,28 @@ class Postgresql(object):
     def check_startup_state_changed(self) -> bool:
         """Checks if PostgreSQL has completed starting up or failed or still starting.
 
-        Should only be called when state == 'starting'
+        Should only be called when state == 'starting [after custom bootstrap]'
 
         :returns: True if state was changed from 'starting'
         """
         ready = self.pg_isready()
 
-        if ready == STATE_REJECT:
+        if ready == PgIsReadyStatus.REJECT:
             return False
-        elif ready == STATE_NO_RESPONSE:
+        elif ready == PgIsReadyStatus.NO_RESPONSE:
             ret = not self.is_running()
             if ret:
-                self.set_state('start failed')
+                self.set_state(PostgresqlState.START_FAILED)
                 self.slots_handler.schedule(False)  # TODO: can remove this?
                 self.config.save_configuration_files(True)  # TODO: maybe remove this?
             return ret
         else:
-            if ready != STATE_RUNNING:
+            if ready != PgIsReadyStatus.RUNNING:
                 # Bad configuration or unexpected OS error. No idea of PostgreSQL status.
                 # Let the main loop of run cycle clean up the mess.
                 logger.warning("%s status returned from pg_isready",
-                               "Unknown" if ready == STATE_UNKNOWN else "Invalid")
-            self.set_state('running')
+                               "Unknown" if ready == PgIsReadyStatus.UNKNOWN else "Invalid")
+            self.set_state(PostgresqlState.RUNNING)
             self.slots_handler.schedule()
             self.config.save_configuration_files(True)
             # TODO: __cb_pending can be None here after PostgreSQL restarts on its own. Do we want to call the callback?
@@ -1000,10 +1019,10 @@ class Postgresql(object):
                 return None
             time.sleep(1)
 
-        return self.state == 'running'
+        return self.state == PostgresqlState.RUNNING
 
     def restart(self, timeout: Optional[float] = None, task: Optional[CriticalTask] = None,
-                block_callbacks: bool = False, role: Optional[str] = None,
+                block_callbacks: bool = False, role: Optional[PostgresqlRole] = None,
                 before_shutdown: Optional[Callable[..., Any]] = None,
                 after_start: Optional[Callable[..., Any]] = None) -> Optional[bool]:
         """Restarts PostgreSQL.
@@ -1013,13 +1032,14 @@ class Postgresql(object):
 
         :returns: True when restart was successful and timeout did not expire when waiting.
         """
-        self.set_state('restarting')
+        self.set_state(PostgresqlState.RESTARTING)
         if not block_callbacks:
             self.__cb_pending = CallbackAction.ON_RESTART
         ret = self.stop(block_callbacks=True, before_shutdown=before_shutdown)\
             and self.start(timeout, task, True, role, after_start)
         if not ret and not self.is_starting():
-            self.set_state('restart failed ({0})'.format(self.state))
+            logger.warning('restart failed (%r)', self.state)
+            self.set_state(PostgresqlState.RESTART_FAILED)
         return ret
 
     def is_healthy(self) -> bool:
@@ -1041,7 +1061,7 @@ class Postgresql(object):
     def controldata(self) -> Dict[str, str]:
         """ return the contents of pg_controldata, or non-True value if pg_controldata call failed """
         # Don't try to call pg_controldata during backup restore
-        if self._version_file_exists() and self.state != 'creating replica':
+        if self._version_file_exists() and self.state != PostgresqlState.CREATING_REPLICA:
             try:
                 env = {**os.environ, 'LANG': 'C', 'LC_ALL': 'C'}
                 data = subprocess.check_output([self.pgcommand('pg_controldata'), self._data_dir], env=env)
@@ -1049,8 +1069,8 @@ class Postgresql(object):
                     data = filter(lambda e: ':' in e, data.decode('utf-8').splitlines())
                     # pg_controldata output depends on major version. Some of parameters are prefixed by 'Current '
                     return {k.replace('Current ', '', 1): v.strip() for k, v in map(lambda e: e.split(':', 1), data)}
-            except subprocess.CalledProcessError:
-                logger.exception("Error when calling pg_controldata")
+            except Exception as e:
+                logger.error("Error when calling pg_controldata: %r", e)
         return {}
 
     def waldump(self, timeline: Union[int, str], lsn: str, limit: int) -> Tuple[Optional[bytes], Optional[bytes]]:
@@ -1110,14 +1130,15 @@ class Postgresql(object):
                 logger.exception('Failed to read and parse %s', (history_path,))
         return history
 
-    def follow(self, member: Union[Leader, Member, None], role: str = 'replica',
+    def follow(self, member: Union[Leader, Member, None], role: PostgresqlRole = PostgresqlRole.REPLICA,
                timeout: Optional[float] = None, do_reload: bool = False) -> Optional[bool]:
         """Reconfigure postgres to follow a new member or use different recovery parameters.
 
         Method may call `on_role_change` callback if role is changing.
 
         :param member: The member to follow
-        :param role: The desired role, normally 'replica', but could also be a 'standby_leader'
+        :param role: The desired role, one of :class:`~misc.PostgresqlRole` values, normally
+            :class:`~misc.PostgresqlRole.REPLICA`, but could also be a :class:`~misc.PostgresqlRole.STANDBY_LEADER`
         :param timeout: start timeout, how long should the `start()` method wait for postgres accepting connections
         :param do_reload: indicates that after updating postgresql.conf we just need to do a reload instead of restart
 
@@ -1135,8 +1156,9 @@ class Postgresql(object):
         # and we know for sure that postgres was already running before, we will only execute on_role_change
         # callback and prevent execution of on_restart/on_start callback.
         # If the role remains the same (replica or standby_leader), we will execute on_start or on_restart
-        change_role = self.cb_called and (self.role in ('primary', 'demoted')
-                                          or not {'standby_leader', 'replica'} - {self.role, role})
+        change_role = self.cb_called and \
+            (self.role in (PostgresqlRole.PRIMARY, PostgresqlRole.DEMOTED)
+             or not {PostgresqlRole.STANDBY_LEADER, PostgresqlRole.REPLICA} - {self.role, role})
         if change_role:
             self.__cb_pending = CallbackAction.NOOP
 
@@ -1161,7 +1183,7 @@ class Postgresql(object):
         for _ in polling_loop(wait_seconds):
             data = self.controldata()
             if data.get('Database cluster state') == 'in production':
-                self.set_role('primary')
+                self.set_role(PostgresqlRole.PRIMARY)
                 return True
 
     def _pre_promote(self) -> bool:
@@ -1196,7 +1218,7 @@ class Postgresql(object):
 
     def promote(self, wait_seconds: int, task: CriticalTask,
                 before_promote: Optional[Callable[..., Any]] = None) -> Optional[bool]:
-        if self.role in ('promoted', 'primary'):
+        if self.role in (PostgresqlRole.PROMOTED, PostgresqlRole.PRIMARY):
             return True
 
         ret = self._pre_promote()
@@ -1220,31 +1242,38 @@ class Postgresql(object):
 
         ret = self.pg_ctl('promote', '-W')
         if ret:
-            self.set_role('promoted')
+            self.set_role(PostgresqlRole.PROMOTED)
             self.call_nowait(CallbackAction.ON_ROLE_CHANGE)
             ret = self._wait_promote(wait_seconds)
         return ret
 
     @staticmethod
     def _wal_position(is_primary: bool, wal_position: int,
-                      received_location: Optional[int], replayed_location: Optional[int]) -> int:
-        return wal_position if is_primary else max(received_location or 0, replayed_location or 0)
+                      receive_lsn: Optional[int], replay_lsn: Optional[int]) -> int:
+        return wal_position if is_primary else max(receive_lsn or 0, replay_lsn or 0)
 
-    def timeline_wal_position(self) -> Tuple[int, int, Optional[int]]:
+    def timeline_wal_position(self) -> Tuple[int, int, Optional[int], Optional[int], Optional[int]]:
+        """Get timeline and various wal positions.
+
+        :returns: a tuple composed of 5 integers representing timeline, ``pg_current_wal_lsn()`` position
+            on primary/the biggest value among receive and replay LSN on replicas, ``pg_control_checkpoint()``
+            value on a standby leader, receive and replay LSN on replicas (if available).
+        """
         # This method could be called from different threads (simultaneously with some other `_query` calls).
         # If it is called not from main thread we will create a new cursor to execute statement.
         if current_thread().ident == self.__thread_ident:
             timeline = self._cluster_info_state_get('timeline') or 0
             wal_position = self._cluster_info_state_get('wal_position') or 0
-            replayed_location = self.replayed_location()
-            received_location = self.received_location()
+            replay_lsn = self.replay_lsn()
+            receive_lsn = self.receive_lsn()
             pg_control_timeline = self._cluster_info_state_get('pg_control_timeline')
         else:
-            timeline, wal_position, replayed_location, received_location, _, pg_control_timeline = \
-                self._query(self.cluster_info_query)[0][:6]
+            timeline, wal_position, replay_lsn, receive_lsn, _, pg_control_timeline, _, write_location = \
+                self._query(self.cluster_info_query)[0][:8]
+            receive_lsn = max(receive_lsn or 0, write_location or 0)
 
-        wal_position = self._wal_position(bool(timeline), wal_position, received_location, replayed_location)
-        return timeline, wal_position, pg_control_timeline
+        wal_position = self._wal_position(bool(timeline), wal_position, receive_lsn, replay_lsn)
+        return timeline, wal_position, pg_control_timeline, receive_lsn, replay_lsn
 
     def postmaster_start_time(self) -> Optional[str]:
         try:
@@ -1255,7 +1284,7 @@ class Postgresql(object):
 
     def last_operation(self) -> int:
         return self._wal_position(self.is_primary(), self._cluster_info_state_get('wal_position') or 0,
-                                  self.received_location(), self.replayed_location())
+                                  self.receive_lsn(), self.replay_lsn())
 
     def configure_server_parameters(self) -> None:
         self._major_version = self.get_major_version()
@@ -1318,7 +1347,7 @@ class Postgresql(object):
                     os.unlink(source)
                     os.symlink(new_name, source)
 
-                new_name = '{0}.{1}'.format(self._data_dir, postfix)
+                new_name = '{0}.{1}'.format(self._data_dir.rstrip(os.sep), postfix)
                 logger.info('renaming data directory to %s', new_name)
                 if os.path.exists(new_name):
                     shutil.rmtree(new_name)
@@ -1327,7 +1356,7 @@ class Postgresql(object):
                 logger.exception("Could not rename data directory %s", self._data_dir)
 
     def remove_data_directory(self) -> None:
-        self.set_role('uninitialized')
+        self.set_role(PostgresqlRole.UNINITIALIZED)
         logger.info('Removing data directory: %s', self._data_dir)
         try:
             if os.path.islink(self._data_dir):
@@ -1365,3 +1394,13 @@ class Postgresql(object):
         self.slots_handler.schedule()
         self.mpp_handler.schedule_cache_rebuild()
         self._sysid = ''
+
+    def _get_gucs(self) -> CaseInsensitiveSet:
+        """Get all available GUCs based on ``postgres --describe-config`` output.
+
+        :returns: all available GUCs in the local Postgres server.
+        """
+        cmd = [self.pgcommand('postgres'), '--describe-config']
+        return CaseInsensitiveSet({
+            line.split('\t')[0] for line in subprocess.check_output(cmd).decode('utf-8').strip().split('\n')
+        })
