@@ -9,6 +9,7 @@ from urllib.parse import urlparse
 from ...dcs import Cluster
 from ...psycopg import connect, ProgrammingError, quote_ident
 from ...utils import parse_int
+from ..misc import PostgresqlRole, PostgresqlState
 from . import AbstractMPP, AbstractMPPHandler
 
 if TYPE_CHECKING:  # pragma: no cover
@@ -75,7 +76,7 @@ class PgDistNode:
 
         :returns: ``True`` if this object represents the ``primary``.
         """
-        return self.role in ('primary', 'demoted')
+        return self.role in (PostgresqlRole.PRIMARY, PostgresqlRole.DEMOTED)
 
     def as_tuple(self, include_nodeid: bool = False) -> Tuple[str, int, str, Optional[int]]:
         """Helper method to compare two :class:`PgDistGroup` objects.
@@ -477,7 +478,8 @@ class CitusHandler(Citus, AbstractMPPHandler, Thread):
         for groupid, worker in cluster.workers.items():
             leader = worker.leader
             if leader and leader.conn_url\
-                    and leader.data.get('role') in ('master', 'primary') and leader.data.get('state') == 'running':
+                    and leader.data.get('role') in (PostgresqlRole.MASTER, PostgresqlRole.PRIMARY)\
+                    and leader.data.get('state') == PostgresqlState.RUNNING:
                 self.add_task('after_promote', groupid, worker, leader.name, leader.conn_url)
 
     def find_task_by_groupid(self, groupid: int) -> Optional[int]:
